/* 
 * video.c  
 * Video handling code for 2D graphics examples
 * (c) 2004, Stuart Dalton / BlackAura (obsidianglow@hotpop.com)
 */

#include "video.h"
#include <png/png.h>
#include <jpeg/jpeg.h>

uint16 *backbuffer_s;
uint32 *backbuffer_l;

/* Macros for image drawing */
#define MIN(a, b) ( (a)<(b)? (a):(b) )
#define MAX(a, b) ( (a)>(b)? (a):(b) )

/* Double Buffering
----------------------------------------------------------------------------- */

/* Flip page */
void dbuffer_flip()
{
	/* Copy the active page pointers to the backbuffer pointers */
	backbuffer_s = vram_s;
	backbuffer_l = vram_l;

	/* Wait for a vertical blank (prevent tearing) */
	vid_waitvbl();

	/* Tell KOS to flip to the next page */
	vid_flip(-1);
}

/* Initialise double buffering */
void dbuffer_init()
{
	int fb_size;

	/* Calculate the size of a single page */
	fb_size = vid_mode->width * vid_mode->height;
	switch(vid_mode->pm)
	{
		case PM_RGB555:
		case PM_RGB565:
			fb_size *= 2;
			break;
		case PM_RGB888:
			fb_size *= 4;
			break;
	}

	/* Set the number of video pages to 2 */
	vid_mode->fb_count = 2;

	/* Set the start positions of pages 0 and 1 */
	vid_mode->fb_base[0] = 0;
	vid_mode->fb_base[1] = fb_size;

	/* Empty VRAM */
	vid_empty();

	/* Do a page flip */
	dbuffer_flip();
}

/* Clear the screen with the specified colour */
void dbuffer_clear(int r, int g, int b)
{
	uint16 pixel16;
	uint32 pixel32;

	switch (vid_mode->pm)
	{
		case PM_RGB555:
			pixel16 = ((r >> 3) << 10)
				| ((g >> 3) << 5)
				| ((b >> 3) << 0);
			sq_set16(backbuffer_s, pixel16, (vid_mode->width * vid_mode->height) * 2);
			break;

		case PM_RGB565:
			pixel16 = ((r >> 3) << 11)
				| ((g >> 2) << 5)
				| ((b >> 3) << 0);
			sq_set16(backbuffer_s, pixel16, (vid_mode->width * vid_mode->height) * 2);
			break;

		case PM_RGB888:
			pixel32 = (r << 16) | (g << 8) | (b << 0);
			sq_set32(backbuffer_l, pixel32, (vid_mode->width * vid_mode->height) * 4);
			break;
	}
}

/* Primitive drawing
----------------------------------------------------------------------------- */
/* Draw a line horizontally across the screen */
void draw_hline(int x1, int x2, int y, uint16 c)
{
	int cx;
	uint16 *line_buffer;

	/* Swap x1 and x2 if necessary */
	if(x1 > x2)
	{
		cx = x1;
		x1 = x2;
		x2 = cx;
	}

	/* Don't bother if the line's off the screen */
	if( (y < 0) || (y >= vid_mode->height) || (x1 >= vid_mode->width) || (x2 < 0) )
		return;

	/* Clip the line to the edge of the screen */
	if(x1 < 0)
		x1 = 0;
	if(x2 >= vid_mode->width)
		x2 = vid_mode->width - 1;

	/* Set line_buffer to the beginning of the line */
	line_buffer = backbuffer_s + x1 + (y * vid_mode->width);

	/* Draw the line */
	for(cx = x1; cx <= x2; cx++)
		*line_buffer++ = c;
}

/* Draw a line vertically up/down the screen */
void draw_vline(int x, int y1, int y2, uint16 c)
{
	int cy;
	uint16 *line_buffer;

	/* Swap y1 and y2 if necessary */
	if(y1 > y2)
	{
		cy = y1;
		y1 = y2;
		y2 = cy;
	}

	/* Don't bother if the line's off the screen */
	if( (x < 0) || (x >= vid_mode->width) || (y1 >= vid_mode->height) || (y2 < 0) )
		return;

	/* Clip the line to the edge of the screen */
	if(y1 < 0)
		y1 = 0;
	if(y2 >= vid_mode->height)
		y2 = vid_mode->height - 1;

	/* Set line_buffer to the beginning of the line */
	line_buffer = backbuffer_s + x + (y1 * vid_mode->width);

	/* Draw the line */
	for(cy = y1; cy <= y2; cy++)
	{
		*line_buffer = c;
		line_buffer += vid_mode->width;
	}
}

/* Draw a line between any two points */
void draw_line(int x1, int y1, int x2, int y2, uint16 c)
{
	int dy = y2 - y1;
	int dx = x2 - x1;
	int stepx, stepy;

	if (dy < 0) { dy = -dy; stepy = -1; } else { stepy = 1; }
	if (dx < 0) { dx = -dx; stepx = -1; } else { stepx = 1; }
	dy <<= 1;
	dx <<= 1;

	DRAW_PIXEL(x1, y1, c);

	if (dx > dy)
	{
		/* Shallow lines */
		int fraction = dy - (dx >> 1);
		while (x1 != x2)
		{
			if (fraction >= 0)
			{
				y1 += stepy;
				fraction -= dx;
			}
			x1 += stepx;
			fraction += dy;
			DRAW_PIXEL(x1, y1, c);
		}
	}
	else
	{
		/* Steep lines */
		int fraction = dx - (dy >> 1);
		while (y1 != y2)
		{
			if (fraction >= 0)
			{
				x1 += stepx;
				fraction -= dy;
			}
			y1 += stepy;
			fraction += dx;
			DRAW_PIXEL(x1, y1, c);
		}
	}
}

/* Draw a hollow box */
void draw_hbox(int x1, int y1, int x2, int y2, uint16 c)
{
	draw_hline(x1, x2, y1, c);
	draw_hline(x1, x2, y2, c);
	draw_vline(x1, y1, y2, c);
	draw_vline(x2, y1, y2, c);
}

/* Draw a filled box */
void draw_fbox(int x1, int y1, int x2, int y2, uint16 c)
{
	int cy;
	for(cy = y1; cy <= y2; cy++)
		draw_hline(x1, x2, cy, c);
}

/* Image loading
----------------------------------------------------------------------------- */
kos_img_t *image_loadpng(char *filename)
{
	kos_img_t *image;
	image = (kos_img_t *)( malloc(sizeof(kos_img_t)) );
	if(!image)
	{
		printf("image_loadpng: Out of memory loading %s\n", filename);
		return NULL;
	}

	if( png_to_img(filename, PNG_NO_ALPHA, image) != 0 )
	{
		printf("image_loadpng: Couldn't load %s\n", filename);
		free(image);
		return NULL;
	}

	return image;
}

kos_img_t *image_loadjpeg(char *filename)
{
	kos_img_t *image;
	image = (kos_img_t *)( malloc(sizeof(kos_img_t)) );
	if(!image)
	{
		printf("image_loadjpeg: Out of memory loading %s\n", filename);
		return NULL;
	}

	if( jpeg_to_img(filename, 1, image) != 0 )
	{ 
		printf("image_loadjpeg: Couldn't load %s\n", filename);
		free(image);
		return NULL;
	}

	return image;
}

void image_free(kos_img_t *image)
{
//	kos_img_free(image, 1);
}

/* Image loading
----------------------------------------------------------------------------- */
void image_draw(kos_img_t *image, int x, int y)
{
	uint16 *src, *src_start;
	uint16 *dst, *dst_start;
	int drawW, drawH;
	int startX, endX;
	int startY, endY;

	if(!image)
		return;

	/* Find the starting coordinates, and abort if off screen */
	startX = MAX(x, 0);
	if(startX >= vid_mode->width)
		return;

	startY = MAX(y, 0);
	if(startY >= vid_mode->height)
		return;

	/* Find the ending coordinates */
	endX = MIN(x + image->w, vid_mode->width);
	if(endX <= 0)
		return;

	endY = MIN(y + image->h, vid_mode->height);
	if(endY <= 0)
		return;

	/* Width, height */
	drawW = endX - startX;
	drawH = endY - startY;

	/* Starting location for the source and destination */
	src_start = (uint16 *)image->data;
	dst_start = backbuffer_s + startX + (startY * vid_mode->width);

	/* If we're off the screen, adjust the start position */
	if(x < 0)
		src_start += -x;
	if(y < 0)
		src_start += (-y * image->w);

	/* Draw the image */
	while(drawH--)
	{
		int dw = drawW;

		/* Copy the current image source, and nudge it down one line */
		src = src_start;
		dst = dst_start;
		src_start += image->w;
		dst_start += vid_mode->width;

		/* Draw the line */
		while(dw--)
			*dst++ = *src++;
	}
}

void image_draw_key(kos_img_t *image, int x, int y, uint16 key)
{
	uint16 *src, *src_start;
	uint16 *dst, *dst_start;
	int drawW, drawH;
	int startX, endX;
	int startY, endY;

	if(!image)
		return;

	/* Find the starting coordinates, and abort if off screen */
	startX = MAX(x, 0);
	if(startX >= vid_mode->width)
		return;

	startY = MAX(y, 0);
	if(startY >= vid_mode->height)
		return;

	/* Find the ending coordinates */
	endX = MIN(x + image->w, vid_mode->width);
	if(endX <= 0)
		return;

	endY = MIN(y + image->h, vid_mode->height);
	if(endY <= 0)
		return;

	/* Width, height */
	drawW = endX - startX;
	drawH = endY - startY;

	/* Starting location for the source and destination */
	src_start = (uint16 *)image->data;
	dst_start = backbuffer_s + startX + (startY * vid_mode->width);

	/* If we're off the screen, adjust the start position */
	if(x < 0)
		src_start += -x;
	if(y < 0)
		src_start += (-y * image->w);

	/* Draw the image */
	while(drawH--)
	{
		int dw = drawW;

		/* Copy the current image source, and nudge it down one line */
		src = src_start;
		dst = dst_start;
		src_start += image->w;
		dst_start += vid_mode->width;

		/* Draw the line */
		while(dw--)
		{
			/* Skip transparent pixels */
			if(*src == key)
			{
				*dst++;
				*src++;
			}
			else
			{
				*dst++ = *src++;
			}
		}
	}
}

void image_draw_key_dbl(kos_img_t *image, int x, int y, uint16 key)
{
	uint16 *src, *src_start;
	uint16 *dst, *dst_start;
	int drawW, drawH;
	int startX, endX;
	int startY, endY;

	if(!image)
		return;

	/* Find the starting coordinates, and abort if off screen */
	startX = MAX(x, 0);
	if(startX >= vid_mode->width)
		return;

	startY = MAX(y, 0);
	if(startY >= vid_mode->height)
		return;

	/* Find the ending coordinates */
	endX = MIN(x + image->w, vid_mode->width);
	if(2*endX <= 0)
		return;

	endY = MIN(y + image->h, vid_mode->height);
	if(2*endY <= 0)
		return;

	/* Width, height */
	drawW = (endX - startX);
	drawH = (endY - startY)-1;

	/* Starting location for the source and destination */
	src_start = (uint16 *)image->data;
	dst_start = backbuffer_s + startX + (startY * vid_mode->width);

	/* If we're off the screen, adjust the start position */
	if(x < 0)
		src_start += -x;
	if(y < 0)
		src_start += (-y * image->w);

	/* Draw the image */
	while(drawH--)
	{
		int dw = drawW;

		/* Copy the current image source, and nudge it down one line */
		src = src_start;
		dst = dst_start;
		
		src_start += image->w;
		dst_start += vid_mode->width;

		/* Draw the line */
		while(dw--)
		{
			/* Skip transparent pixels */
			if(*src == key)
			{
				*dst++;
				*dst++;
				*src++;
			}
			else
			{
				*dst++ = *src;
				*dst++ = *src++;
			}
		}
		dw = drawW;

		/* Copy the current image source, and nudge it down one line */
		src = src_start;
		dst = dst_start;
		dst_start += vid_mode->width;

		/* Draw the line */
		while(dw--)
		{
			/* Skip transparent pixels */
			if(*src == key)
			{
				*dst++;
				*dst++;
				*src++;
			}
			else
			{
				*dst++ = *src;
				*dst++ = *src++;
			}
		}
	}
}
